import 'package:dio/dio.dart' as dio;

import 'http.dart';
import 'http_error.dart';
import 'options.dart';
import 'response.dart';

abstract class FHttpMixin implements FHttp {
  @override
  late FHttpBaseOptions options;

  @override
  late final dio.HttpClientAdapter httpClientAdapter;

  dio.Dio? _dioInstance;

  bool _closed = false;

  @override
  void close({bool force = false}) {
    if (!_closed) {
      _closed = true;
      _dioInstance?.close(force: force);
    }
  }

  @override
  Future<FHttpResponse<T>> get<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    FHttpScopedOptions? options,
    FHttpCancelToken? cancelToken,
    dio.ProgressCallback? onReceiveProgress,
  }) async {
    try {
      _checkProxyInstance();
      _loadScopedOptions(options);

      dio.Response<T> dioRsp = await _dioInstance!.get<T>(
        path,
        queryParameters: queryParameters,
        options: _map2DioOptions(options),
        cancelToken: cancelToken,
        onReceiveProgress: onReceiveProgress,
      );
      return _mapFromDioResponse(dioRsp);
    } on dio.DioError catch (e) {
      throw _mapFromDioError(e);
    }
  }

  @override
  Future<FHttpResponse<T>> post<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    FHttpScopedOptions? options,
    FHttpCancelToken? cancelToken,
    dio.ProgressCallback? onSendProgress,
    dio.ProgressCallback? onReceiveProgress,
  }) async {
    try {
      _checkProxyInstance();
      _loadScopedOptions(options);

      dio.Response<T> dioRsp = await _dioInstance!.post<T>(
        path,
        data: data,
        queryParameters: queryParameters,
        options: _map2DioOptions(options),
        cancelToken: cancelToken,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
      return _mapFromDioResponse(dioRsp);
    } on dio.DioError catch (e) {
      throw _mapFromDioError(e);
    }
  }

  @override
  Future<FHttpResponse<T>> put<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    FHttpScopedOptions? options,
    FHttpCancelToken? cancelToken,
    dio.ProgressCallback? onSendProgress,
    dio.ProgressCallback? onReceiveProgress,
  }) async {
    try {
      _checkProxyInstance();
      _loadScopedOptions(options);

      dio.Response<T> dioRsp = await _dioInstance!.put<T>(
        path,
        data: data,
        queryParameters: queryParameters,
        options: _map2DioOptions(options),
        cancelToken: cancelToken,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
      return _mapFromDioResponse(dioRsp);
    } on dio.DioError catch (e) {
      throw _mapFromDioError(e);
    }
  }

  @override
  Future<FHttpResponse<T>> delete<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    FHttpScopedOptions? options,
    FHttpCancelToken? cancelToken,
  }) async {
    try {
      _checkProxyInstance();
      _loadScopedOptions(options);

      dio.Response<T> dioRsp = await _dioInstance!.delete<T>(path,
          queryParameters: queryParameters,
          options: _map2DioOptions(options),
          cancelToken: cancelToken);
      return _mapFromDioResponse(dioRsp);
    } on dio.DioError catch (e) {
      throw _mapFromDioError(e);
    }
  }

  void _checkProxyInstance() {
    if (_dioInstance == null) {
      _dioInstance = new dio.Dio(_map2DioBaseOptions(options))
        ..httpClientAdapter = this.httpClientAdapter;
    }
  }

  void _loadScopedOptions(FHttpScopedOptions? options) {
    final ins = _dioInstance!;
    if (options == null) return;
    if (options.responseType != null) {
      ins.options.responseType =
          dio.ResponseType.values[options.responseType!.index];
    }
    if (options.contentType != null) {
      ins.options.contentType = options.contentType!;
    }
    if (options.timeout != null && options.timeout! > 0) {
      ins.options.connectTimeout = options.timeout!;
    } else {
      ins.options.connectTimeout = this.options.timeout ?? 0;
    }
  }
}

dio.BaseOptions? _map2DioBaseOptions(FHttpBaseOptions? options) {
  if (options == null) return null;
  return dio.BaseOptions(
    baseUrl: options.baseUrl,
    connectTimeout: options.timeout,
    method: options.method,
    queryParameters: options.queryParameters,
    headers: options.headers,
    responseType: _map2DioResponseType(options.responseType),
    contentType: options.contentType,
  );
}

dio.Options? _map2DioOptions(FHttpScopedOptions? options) {
  if (options == null) return null;
  return dio.Options(
    method: options.method,
    headers: options.headers,
    responseType: _map2DioResponseType(options.responseType),
    contentType: options.contentType,
  );
}

FHttpResponse<T> _mapFromDioResponse<T>(dio.Response<T> dioRsp) {
  var response = FHttpResponse<T>(
    data: dioRsp.data,
    statusCode: dioRsp.statusCode,
    statusMessage: dioRsp.statusMessage,
    headers: dioRsp.headers.map,
  );
  return response;
}

FHttpError _mapFromDioError(dio.DioError e) {
  FHttpErrorType type = _maoFromDioErrorType(e.type);
  FHttpResponse? response;
  if (e.response != null) {
    response = _mapFromDioResponse(e.response!);
  }
  return FHttpError(response: response, type: type, error: e.error);
}

FHttpErrorType _maoFromDioErrorType(dio.DioErrorType dioType) {
  switch (dioType) {
    case dio.DioErrorType.connectTimeout:
      return FHttpErrorType.connectTimeout;
    case dio.DioErrorType.response:
      return FHttpErrorType.response;
    case dio.DioErrorType.cancel:
      return FHttpErrorType.cancel;
    case dio.DioErrorType.sendTimeout:
    case dio.DioErrorType.receiveTimeout:
    case dio.DioErrorType.other:
      return FHttpErrorType.cancel;
  }
}

dio.ResponseType? _map2DioResponseType(FHttpResponseType? type) {
  if (type == null) return null;
  return dio.ResponseType.values[type.index];
}
