import 'logger/logger.dart';
import 'package:flutter/foundation.dart';

import 'extensions/type.dart';

typedef T FJsonDecoder<T>(dynamic data);
typedef Map<String, dynamic> FJsonEncoder<T>();

class FJsonConvert {
  FJsonConvert._();

  static Map<Type, FJsonDecoder> _decoders = {
    typeOf<List<String>>(): (map) => (map as List).cast<String>(),
    typeOf<List<int>>(): (map) => (map as List).cast<int>(),
    typeOf<List<double>>(): (map) => (map as List).cast<double>(),
    typeOf<List<bool>>(): (map) => (map as List).cast<bool>(),
    typeOf<DateTime>(): (val) => DateTime.parse(val),
  };

  /// 序列化到Map
  static Map<String, dynamic> toJson<T extends Object>(T? data) {
    if (data == null) return {};
    try {
      final fnToJson = (data as dynamic)['toJson'];
      if (fnToJson == null && fnToJson is FJsonEncoder) {
        final map = fnToJson.call();
        return map;
      }
    } catch (e) {
      logger.e("FJsonConvert handle toJson error.", e);
    }
    return {};
  }

  /// 反序列化到实体
  static T? fromJson<T>(dynamic data) {
    if (data == null) return null;
    try {
      if (data is Map || data is List) {
        return _decodeToModel<T>(data);
      }
      return data as T;
    } catch (e) {
      logger.e("FJsonConvert handle fromJson error.", e);
    }
    return null;
  }

  /// 反序列化到列表
  static List<T?> fromListJson<T>(dynamic data) {
    if (data == null) return [];
    try {
      if (data is List) {
        List<dynamic> dataList = data;
        if (dataList.isNotEmpty) {
          return dataList.map<T?>((e) => fromJson<T>(e)).toList();
        }
      }
      return [fromJson<T>(data)];
    } catch (e) {
      logger.e("FJsonConvert handle fromListJson error.", e);
    }
    return [];
  }

  /// 设置实体的json解码器(反序列化设置)
  static void setDecoder<T>(FJsonDecoder<T> decoder) {
    var type = typeOf<T>();
    _decoders[type] = decoder;
    setListDecoder(decoder);
  }

  /// 设置List的json解码器(反序列化设置)【用setDecoder会自动注册对应list解码器】
  static void setListDecoder<T>(FJsonDecoder<T> decoder) {
    var type = typeOf<List<T>>();
    _decoders[type] = (list) => (list as List).map((e) => decoder(e)).toList();
  }

  static _decodeToModel<T>(map) {
    var type = typeOf<T>();
    if (!_decoders.containsKey(type))
      throw FlutterError("The FJsonConvert-Decoder for $type is not found!");

    var model = _decoders[type]!.call(map);
    return model;
  }
}
