import '../env/env.dart';
import 'package:flutter/foundation.dart';
import 'package:intl/intl.dart';
import 'logger_stub.dart'
    if (dart.library.io) 'logger_io.dart'
    if (dart.library.js) 'logger_web.dart';

enum LogLevel {
  Verbose,
  Degbug,
  Info,
  Warn,
  Error,
  UserOperate,
}

abstract class Logger {
  static const Map<LogLevel, String> _logLevelTags = {
    LogLevel.Verbose: "V",
    LogLevel.Degbug: "D",
    LogLevel.Info: "I",
    LogLevel.Warn: "W",
    LogLevel.Error: "E",
    LogLevel.UserOperate: "U",
  };

  LogLevel _minimumLevelLimit = LogLevel.Info;

  /// 设置最小日志等级(低于该等级的记录自动忽略)
  void setMinimumLevel(LogLevel level) => _minimumLevelLimit = level;

  /// 详细日志
  void v(String msg) => _handleLog(LogLevel.Verbose, msg);

  /// 调试日志
  void d(String msg) => _handleLog(LogLevel.Degbug, msg);

  /// 信息日志
  void i(String msg) => _handleLog(LogLevel.Info, msg);

  /// 警告日志
  void w(String msg) => _handleLog(LogLevel.Warn, msg);

  /// 异常日志
  ///
  /// [e] 应尽量使用[Error]或[Exception]类型
  void e(String msg, [Object? e]) {
    String? eStr = getExceptionString(e);
    String content = eStr != null && eStr.isNotEmpty ? "$msg - $eStr" : msg;
    _handleLog(LogLevel.Error, content);
  }

  /// 记录用户日志
  ///
  /// [accountName] 用户名
  void u(String msg, [String accountName = '']) {
    final uMsg = "$accountName:-> $msg";
    _handleLog(LogLevel.UserOperate, uMsg);
  }

  void _handleLog(LogLevel level, String msg) {
    if (_minimumLevelLimit.index <= level.index) {
      Future.microtask(() => write(level, msg));
    }
  }

  /// 写一条日志
  @protected
  Future<void> write(LogLevel level, String msg);

  /// 组装日志内容
  @protected
  String buildLogContent(LogLevel level, String msg) {
    var levelTag = getLevelTag(level);
    var timeStr = getLogTimeString();
    return "[$timeStr]-$levelTag; $msg";
  }

  @protected
  String? getExceptionString(Object? e) {
    if (e == null) return null;
    if (e is Error && e.stackTrace != null) {
      return '${e.toString()}: ${e.stackTrace.toString()}';
    }
    return e.toString();
  }

  @protected
  String getLogTimeString() =>
      DateFormat('yyyyMMddHHmmss,SSS').format(DateTime.now());

  @protected
  String getLevelTag(LogLevel level) {
    if (_logLevelTags.containsKey(level)) {
      return _logLevelTags[level]!;
    } else {
      return 'Unknown';
    }
  }

  /// 初始化
  Future<void> init() async {}

  /// 释放
  Future<void> dispose() async {}
}

Logger _buildLogger() {
  switch (FPlatform.current) {
    case FPlatformEnum.android:
    case FPlatformEnum.iOS:
      return NativeLogger();
    case FPlatformEnum.web:
      return WebLogger();
    case FPlatformEnum.webOnWin:
    case FPlatformEnum.webOnMac:
      return WebShellLogger();
  }
}

final logger = _buildLogger();
